/*
 *  cam.c
 *
 * Test program for video capture and tv output,
 * using the Video4Linux driver.
 * Original by Pat Beirne, rewritten by Russell King
 *
 *  Copyright (C) 1995-2000 Russell King.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
#include <stdio.h>
#include <sys/types.h>
#include <linux/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <time.h>
#include <stdlib.h>
#include <assert.h>
#include <unistd.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <zlib.h>

#include <vga.h>
#include <linux/videodev.h>
#include <linux/fb.h>

#include "yuvtorgb.h"

int			vid_dev;
struct video_capability cap;
struct video_window	win;
struct video_picture	pic;

struct fb_var_screeninfo fb_var;
struct fb_fix_screeninfo fb_fix;

static int get_screen_params(void)
{
	int fd;

	fd = open("/dev/fb", O_RDONLY | O_NONBLOCK);
	if (fd < 0) {
		fprintf(stderr, "Unable to open /dev/fb: %s\n",
			strerror(errno));
		return -1;
	}

	if (ioctl(fd, FBIOGET_FSCREENINFO, &fb_fix) == -1) {
		fprintf(stderr, "GETFSCREENINFO failed: %s\n",
			strerror(errno));
		close(fd);
		return -1;
	}

	if (ioctl(fd, FBIOGET_VSCREENINFO, &fb_var) == -1) {
		fprintf(stderr, "GETVSCREENINFO failed: %s\n",
			strerror(errno));
		close(fd);
		return -1;
	}
	close(fd);

	return 0;
}

static int open_vidcap(char *dev)
{
	struct video_buffer buf;
	int fd;
	int r;

	/*
	 * Open video capture device
	 */
	fd = open(dev, O_RDONLY | O_NONBLOCK);

	if (fd < 0) {
		fprintf(stderr, "Unable to open %s: %s\n",
			dev, strerror(errno));

		if (errno == ENODEV)
			fprintf(stderr,
				"  please ensure that the video capture "
				"driver is installed\n");
		if (errno == EACCES)
			fprintf(stderr,
				"  please check permissions on %s; perhaps "
				"you need to be root\n", dev);
		exit(1);
	}

	/*
	 * Get capabilities
	 */
	r = ioctl(fd, VIDIOCGCAP, &cap);
	if (r != 0) {
		fprintf(stderr, "VIDIOCGCAP failed: %s\n", strerror(errno));
		close(fd);
		return -1;
	}

	/*
	 * Tell the video capture driver about
	 * our framebuffer
	 */
	buf.base         = (void *)0;	/* video memory located at start */
	buf.width        = fb_var.xres_virtual;
	buf.height       = fb_var.yres_virtual;
	buf.depth        = fb_var.bits_per_pixel;
	buf.bytesperline = buf.width * buf.depth / 8;

	r = ioctl(fd, VIDIOCSFBUF, &buf);
	if (r != 0) {
		fprintf(stderr, "VIDIOCSFBUF failed: %s\n", strerror(errno));
		close(fd);
		return -1;
	}

	/*
	 * Get current settings
	 */
	r = ioctl(fd, VIDIOCGWIN, &win);
	if (r != 0) {
		fprintf(stderr, "VIDIOCGWIN failed: %s\n", strerror(errno));
		close(fd);
		return -1;
	}

	/*
	 * Get current picture settings
	 */
	r = ioctl(fd, VIDIOCGPICT, &pic);
	if (r != 0) {
		fprintf(stderr, "VIDIOCGPICT failed: %s\n", strerror(errno));
		close(fd);
 		return -1;
	}
	
	return fd;
}

static int overlay_set_win(int x, int y, int w, int h)
{
	int r;

	if (h > cap.maxheight)
		h = cap.maxheight;
	else if (h < cap.minheight)
		h = cap.minheight;

	if (w > cap.maxwidth)
		w = cap.maxwidth;
	else if (w < cap.minwidth)
		w = cap.minwidth;

	win.x      = x;
	win.y      = y;
	win.width  = w;
	win.height = h;
	win.flags  = VIDEO_WINDOW_INTERLACE;

	r = ioctl(vid_dev, VIDIOCSWIN, &win);
	if (r != 0)
		fprintf(stderr, "VIDIOCSWIN failed: %s\n", strerror(errno));

	return r;
}

static int update_pic(int vid_dev, struct video_picture *pic)
{
	int r;

	r = ioctl(vid_dev, VIDIOCSPICT, pic);
	if (r)
		fprintf(stderr, "VIDIOCSPICT failed: %s\n", strerror(errno));

	return r;
}

static int init_capture(int vid_dev)
{
	win.x = 0;
	win.y = 0;
	win.width  = cap.maxwidth;
	win.height = cap.maxheight;

	if (win.width > fb_var.xres)
		win.width = fb_var.xres;
	if (win.height > fb_var.yres)
		win.height = fb_var.yres;

	win.x = (fb_var.xres - win.width) / 2;
	win.y = (fb_var.yres - win.height) / 2;

	overlay_set_win(win.x, win.y, win.width, win.height);
	update_pic(vid_dev, &pic);

	return 0;
}

static int overlay(int vid_dev, int on_off)
{
	int r;

	r = ioctl(vid_dev, VIDIOCCAPTURE, &on_off);
	if (r)
		fprintf(stderr, "VIDIOCCAPTURE failed: %s\n", strerror(errno));

	return r;
}

void stop_cam(void)
{
	overlay(vid_dev, 0);

	if (vid_dev >= 0)
		close(vid_dev);
}

struct grabbuf {
	int	width;
	int	height;
	int	size;
	void	*buffer;
};

#define grab_buffer(gbuf)	((gbuf)->buffer)
#define grab_size(gbuf)		((gbuf)->size)
#define grab_width(gbuf)	((gbuf)->width)
#define grab_height(gbuf)	((gbuf)->height)

static struct grabbuf *grab_alloc(void)
{
	struct grabbuf *buf;
	int size;

	size = cap.maxwidth * cap.maxheight * 2 + sizeof(struct grabbuf);

	buf = malloc(size);
	if (buf) {
		buf->buffer = buf + 1;
		buf->width = cap.maxwidth;
		buf->height = cap.maxheight;
		buf->size = size - sizeof(struct grabbuf);
	}
	return buf;
}

static int grab_image(struct grabbuf *gbuf)
{
	int len;

	len = read(vid_dev, gbuf->buffer, gbuf->size);
	if (len == -1) {
		perror("grab read");
		return -1;
	}

	return 0;
}

static int timediff(struct timeval *tv2, struct timeval *tv1)
{
	int diff;

	diff = (tv2->tv_sec - tv1->tv_sec) * 1000;
	diff += (tv2->tv_usec - tv1->tv_usec) / 1000;

	return diff;
}

static void *create_ppm(char *fn, int size, int width, int height)
{
	char str[32];
	char *buf;
	int fd, sz;

	sz = sprintf(str, "P6\n0%d %d\n%d\n", width, height, 255);

	fd = open(fn, O_RDWR|O_CREAT, 0644);
	if (fd == -1) {
		perror("open");
		return NULL;
	}

	write (fd, str, sz);

	lseek(fd, size, SEEK_CUR);
	write(fd, "\0", 1);

	buf = mmap(NULL, size, PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0);
	if (buf == (char *)-1) {
		perror("mmap");
		close(fd);
		return NULL;
	}
	close(fd);

	return buf + sz;
}

int main(int argc, char **argv)
{
	struct grabbuf *gbuf;
	unsigned int ppmsize;
	void *ppmbuf;

	yuv422_to_rgb_init();

	if (get_screen_params())
		exit(1);

	vid_dev = open_vidcap(VIDEO_DEV);
	if (vid_dev < 0)
		exit(1);

	gbuf = grab_alloc();
	if (!gbuf) {
		perror("grab_alloc");
		exit(1);
	}

	ppmsize = grab_size(gbuf) * 6 / 4;
	ppmbuf = create_ppm(CAM_PPM_FILE, ppmsize, grab_width(gbuf), grab_height(gbuf));
	if (!ppmbuf) {
		perror("create_ppm");
		exit(1);
	}

	init_capture(vid_dev);
	atexit(stop_cam);
	overlay(vid_dev, 1);

	while (1) {
		struct timeval start, gdone, ydone, edone;

		printf("grabbing.");
		fflush(stdout);

		gettimeofday(&start, NULL);
		grab_image(gbuf);
		printf(".");
		fflush(stdout);

		gettimeofday(&gdone, NULL);
		yuv422_to_rgb24(ppmbuf, grab_buffer(gbuf), grab_size(gbuf));
		printf(".");
		fflush(stdout);

		gettimeofday(&ydone, NULL);
		system(CAM_SH " " CAM_PPM_FILE);
		printf(".");
		fflush(stdout);

		gettimeofday(&edone, NULL);
		printf("done (grab %03d yuv %03d encode %d)\n",
			timediff(&gdone, &start),
			timediff(&ydone, &gdone),
			timediff(&edone, &ydone));
	}
	return 0;
}
